open SysPath_type;;

exception SysPathBaseFilenameRelative;;
exception SysPathRelativeUnreducable;;
exception SysPathUnrecognizedOS of string;;
exception SysPathFilenameMultiple;;
exception SysPathEmpty;;
exception SysPathNoExtension;;
exception SysPathInvalidFilename;;

type filename = SysPath_type.filename
;;

type filename_part = SysPath_type.filename_part
;;

type extension = SysPath_type.extension
;;

module type OS_SPECIFICATION =
sig
	val dir_writer                : (filename_part list) -> string
	val dir_reader                : Lexing.lexbuf -> (filename_part list)
	val path_writer               : (filename list) -> string
	val path_reader               : Lexing.lexbuf -> (filename list)
end
;;

module type PATH_SPECIFICATION =
sig
val basename        : filename -> filename
val dirname         : filename -> filename
val up_dir          : filename -> filename 
val concat      : filename -> filename -> filename
val make_filename : filename list -> filename 
val reduce : filename -> filename
val make_absolute : filename -> filename -> filename
val make_relative : filename -> filename -> filename
val reparent : filename -> filename -> filename -> filename 
val identity : filename -> filename
val is_valid : filename -> bool
val is_relative : filename -> bool
val is_implicit : filename -> bool
val chop_extension  : filename -> filename
val get_extension   : filename -> extension 
val check_extension : filename -> extension -> bool
val add_extension   : filename -> extension -> filename
val make_path_variable : filename list -> string
val read_path_variable : string -> filename list
val current_dir : filename
val parent_dir  : filename
val root        : string -> filename
val component   : string -> filename
val implode : filename_part list -> filename
val explode : filename -> filename_part list   
end
;;

module type META_PATH_SPECIFICATION =
functor ( OsOperation : OS_SPECIFICATION ) -> 
PATH_SPECIFICATION
;;

module GenericPath : META_PATH_SPECIFICATION = 
functor ( OsOperation : OS_SPECIFICATION ) ->
struct
	(* Debug function *)

	let debug_print_component lst=
		let debug_print_one_component elem =
			match elem with
			  Root s      -> "Root : "^s
			| Component s -> "Component : "^s
			| ParentDir   -> "ParenDir"
			| CurrentDir  -> "CurrentDir"
		in
		List.iter print_string (List.map (fun x -> (debug_print_one_component x)^" ;") lst);
		print_newline ()

	(* Explode *)

	let explode str = 
		try 
			let lexbuf = Lexing.from_string str
			in
			OsOperation.dir_reader lexbuf
		with Parsing.Parse_error ->
			raise SysPathInvalidFilename

	(* Implode *)

	let implode lst = 
		OsOperation.dir_writer lst 

	(* Concat *)

	let concat_list lst_fln1 lst_fln2 =
		implode (lst_fln1 @ lst_fln2)
	
	let concat fln1 fln2 = 
		concat_list  (explode fln1) (explode fln2)

	(* Is_relative *)

	let is_relative_list lst_fln =
		match lst_fln with
		 (Root _) :: _ -> false
		| _            -> true

	let is_relative fln  = 
		is_relative_list ( explode fln )
	
	(* Is_implicit *)
	
	let is_implicit_list lst_fln  = 
		match lst_fln with
		  ParentDir :: _ 
		| CurrentDir :: _ 
		| Component _ :: _ -> true
		| _                -> false

	let is_implicit fln =
		is_implicit_list ( explode fln )

	
	(* Is_valid *)
	
	let is_valid fln =
		try
			let _ = explode fln
			in
			true
		with SysPathInvalidFilename ->
			false

	(* Basename *)

	let basename fln = 
		match List.rev ( explode fln ) with	
		  hd :: tl ->
			implode ( [hd] )
		| [] ->
			raise SysPathEmpty

	(* Dirname *)

	let dirname fln = 
		match List.rev ( explode fln ) with
		  hd :: tl ->
			implode (List.rev tl)
		| [] ->
			raise SysPathEmpty

	(* Extension manipulation *)

	let split_extension fln = 
		match explode (basename fln) with
		  (Component str) :: []->
			let lexbuf = Lexing.from_string str
			in
			let (base,ext) = GenericPath_parser.main_extension
				GenericPath_lexer.token_extension
				lexbuf
			in
			(concat (dirname fln) base, ext)
		| _ ->
			raise SysPathNoExtension

	let check_extension fln ext = 
		let (real_fln, real_ext) = split_extension fln
		in
		ext = real_ext 

	let get_extension fln = 
		let (real_fln, real_ext) = split_extension fln
		in
		real_ext

	let chop_extension  fln =
		let (real_fln, real_ext) = split_extension fln
		in
		real_fln

	let add_extension fln ext =
		fln^"."^ext

	(* Reduce *)

	let rec reduce_list path_lst =
		let stack_dir = Stack.create ()
		in
		let to_list () =
			let tmp_arr = Array.make (Stack.length stack_dir) CurrentDir 
			in
			for i = (Stack.length stack_dir) - 1 downto 0 do 
				Array.set tmp_arr i (Stack.pop stack_dir) 
			done;
			Array.to_list tmp_arr
		in
		let walk_path itm =
			match itm with
			  ParentDir    -> 
			  	begin
			  	try
					begin
					match Stack.pop stack_dir with
					  Root s    -> 
					  	Stack.push (Root s) stack_dir
					| ParentDir -> 
						Stack.push ParentDir stack_dir;
						Stack.push ParentDir stack_dir
					| _         -> 
						()
					end
				with Stack.Empty ->
					Stack.push ParentDir stack_dir
				end
			| CurrentDir   -> 
				()
			| Component "" -> 
				()
			| Component _ 
			| Root _       -> 
				Stack.push itm stack_dir
		in
		let lst = 
			List.iter walk_path path_lst;
			to_list ()
		in
		lst

	let reduce path =
		implode (reduce_list (explode path))

	(* Make_asbolute *)

	let make_absolute_list lst_base lst_path =
		if is_relative_list lst_base then
			raise SysPathBaseFilenameRelative
		else if is_relative_list lst_path then
			reduce_list (lst_base @ lst_path)
		else
			reduce_list (lst_path)


	let make_absolute base_path path =
		implode (make_absolute_list (explode base_path) (explode path))

	(* Make_relative *)

	let make_relative_list lst_base lst_path =
		let rec make_relative_list_aux lst_base lst_path =
			match  (lst_base, lst_path) with
			x :: tl_base, a :: tl_path when x = a ->
				make_relative_list_aux tl_base tl_path
			| _, _ ->
				let back_to_base = List.rev_map 
					(fun x -> ParentDir)
					lst_base
				in
				back_to_base @ lst_path
		in
		if is_relative_list lst_base then
			raise SysPathBaseFilenameRelative
		else if is_relative_list lst_path then
			(reduce_list lst_path)
		else
			make_relative_list_aux (reduce_list lst_base) (reduce_list lst_path)


	let make_relative base_path path =
		implode (make_relative_list (explode base_path) (explode path))

	(* Make_filename *)

	let make_filename_list lst_lst =
		List.flatten lst_lst
		
	let make_filename lst =
		implode ( make_filename_list ( List.map explode lst ) )

	(* Reparent *)

	let reparent_list lst_src lst_dst lst_fln =
		let lst_relative =
			make_relative_list lst_src lst_fln
		in
		make_absolute_list lst_dst lst_relative

	let reparent fln_src fln_dst fln =
		implode (reparent_list (explode fln_src) (explode fln_dst) (explode fln))

	(* Identity *)
	
	let identity fln =
		implode (explode fln)

	
	(* Dangerous functions *)

	let parent_dir  = implode [ParentDir]

	let current_dir = implode [CurrentDir]

	let root s      = implode [Root s]

	let component s = implode [Component s]

	let up_dir fln  = reduce ( concat fln parent_dir )

	(* Manipulate path like variable *)

	let make_path_variable lst = 
		OsOperation.path_writer lst

	let read_path_variable str = 
		try
			let lexbuf = Lexing.from_string str
			in
			OsOperation.path_reader lexbuf
		with Parsing.Parse_error ->
			raise SysPathInvalidFilename
end 
;;


module UnixPath : PATH_SPECIFICATION = GenericPath(struct
	let dir_writer                = UnixPath.dir_writer
	let dir_reader                = UnixPath.dir_reader
	let path_writer               = UnixPath.path_writer
	let path_reader               = UnixPath.path_reader
end)
;;

module MacOSPath : PATH_SPECIFICATION = GenericPath(struct
	let dir_writer                = MacOSPath.dir_writer
	let dir_reader                = MacOSPath.dir_reader
	let path_writer               = MacOSPath.path_writer
	let path_reader               = MacOSPath.path_reader
end)
;;

module Win32Path : PATH_SPECIFICATION = GenericPath(struct 
	let dir_writer                = Win32Path.dir_writer
	let dir_reader                = Win32Path.dir_reader
	let path_writer               = Win32Path.path_writer
	let path_reader               = Win32Path.path_reader
end)
;;

module CygwinPath : PATH_SPECIFICATION = GenericPath(struct
	let dir_writer                = CygwinPath.dir_writer
	let dir_reader                = CygwinPath.dir_reader
	let path_writer               = CygwinPath.path_writer
	let path_reader               = CygwinPath.path_reader
end)
;;

let
(
 basename,       dirname,            up_dir, 
 concat,         reduce,             make_absolute, 
 make_relative,  reparent,           identity,           
 is_valid,       is_relative,        is_implicit,
 chop_extension, get_extension,      check_extension, 
 add_extension,  make_path_variable, read_path_variable, 
 current_dir,    parent_dir,         root, 
 component,      implode,            explode,
 make_filename
)
=
	match Sys.os_type with
	  "Unix" ->
(
 UnixPath.basename,       UnixPath.dirname,            UnixPath.up_dir, 
 UnixPath.concat,         UnixPath.reduce,             UnixPath.make_absolute, 
 UnixPath.make_relative,  UnixPath.reparent,           UnixPath.identity,           
 UnixPath.is_valid,       UnixPath.is_relative,        UnixPath.is_implicit,
 UnixPath.chop_extension, UnixPath.get_extension,      UnixPath.check_extension, 
 UnixPath.add_extension,  UnixPath.make_path_variable, UnixPath.read_path_variable, 
 UnixPath.current_dir,    UnixPath.parent_dir,         UnixPath.root, 
 UnixPath.component,      UnixPath.implode,            UnixPath.explode,
 UnixPath.make_filename
)
	| "MacOS" ->
(
 MacOSPath.basename,       MacOSPath.dirname,            MacOSPath.up_dir, 
 MacOSPath.concat,         MacOSPath.reduce,             MacOSPath.make_absolute, 
 MacOSPath.make_relative,  MacOSPath.reparent,           MacOSPath.identity,           
 MacOSPath.is_valid,       MacOSPath.is_relative,        MacOSPath.is_implicit,
 MacOSPath.chop_extension, MacOSPath.get_extension,      MacOSPath.check_extension, 
 MacOSPath.add_extension,  MacOSPath.make_path_variable, MacOSPath.read_path_variable, 
 MacOSPath.current_dir,    MacOSPath.parent_dir,         MacOSPath.root, 
 MacOSPath.component,      MacOSPath.implode,            MacOSPath.explode,
 MacOSPath.make_filename
)
	| "Win32" ->
(
 Win32Path.basename,       Win32Path.dirname,            Win32Path.up_dir, 
 Win32Path.concat,         Win32Path.reduce,             Win32Path.make_absolute, 
 Win32Path.make_relative,  Win32Path.reparent,           Win32Path.identity,           
 Win32Path.is_valid,       Win32Path.is_relative,        Win32Path.is_implicit,
 Win32Path.chop_extension, Win32Path.get_extension,      Win32Path.check_extension, 
 Win32Path.add_extension,  Win32Path.make_path_variable, Win32Path.read_path_variable, 
 Win32Path.current_dir,    Win32Path.parent_dir,         Win32Path.root, 
 Win32Path.component,      Win32Path.implode,            Win32Path.explode,
 Win32Path.make_filename
)
	| "Cygwin" ->
(
 CygwinPath.basename,       CygwinPath.dirname,            CygwinPath.up_dir, 
 CygwinPath.concat,         CygwinPath.reduce,             CygwinPath.make_absolute, 
 CygwinPath.make_relative,  CygwinPath.reparent,           CygwinPath.identity,           
 CygwinPath.is_valid,       CygwinPath.is_relative,        CygwinPath.is_implicit,
 CygwinPath.chop_extension, CygwinPath.get_extension,      CygwinPath.check_extension, 
 CygwinPath.add_extension,  CygwinPath.make_path_variable, CygwinPath.read_path_variable, 
 CygwinPath.current_dir,    CygwinPath.parent_dir,         CygwinPath.root, 
 CygwinPath.component,      CygwinPath.implode,            CygwinPath.explode,
 CygwinPath.make_filename
)
	| s ->
		raise (SysPathUnrecognizedOS s)
;;
